# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/CLI.ipynb.

# %% auto 0
__all__ = ['logger', 'app', 'generate_fastkafka_app']

# %% ../nbs/CLI.ipynb 1
from typing import *

import typer
from pathlib import Path
import shutil

from ._components.logger import get_logger
from faststream_gen._code_generator.app_description_validator import (
    validate_app_description,
)
from faststream_gen._code_generator.helper import (
    set_logger_level,
    add_tokens_usage,
    get_relevant_prompt_examples,
    strip_white_spaces,
    write_file_contents,
    ensure_openai_api_key_set,
)
from ._code_generator.constants import MODEL_PRICING, EMPTY_DESCRIPTION_ERROR, OpenAIModel, LOGS_DIR_NAME, INCOMPLETE_APP_ERROR_MSG
from ._components.new_project_generator import create_project
from ._code_generator.app_skeleton_generator import generate_app_skeleton
from ._code_generator.app_and_test_generator import generate_app_and_test
from ._components.integration_test_generator import fix_requirements_and_run_tests

# %% ../nbs/CLI.ipynb 3
logger = get_logger(__name__)

# %% ../nbs/CLI.ipynb 6
app = typer.Typer(
    short_help="Commands for accelerating FastStream project creation using advanced AI technology",
     help="""Commands for accelerating FastStream project creation using advanced AI technology.

These commands use OpenAI's model to generate FastStream project. To access this feature, kindly sign up if you haven't already and create an API key with OpenAI. You can generate API keys in the OpenAI web interface. See https://platform.openai.com/account/api-keys for details.

Once you have the key, please set it in the OPENAI_API_KEY environment variable before executing the code generation commands.

Note: Accessing OpenAI API incurs charges. For further information on pricing and free credicts, check this link: https://openai.com/pricing
    """,
)

# %% ../nbs/CLI.ipynb 7
def _calculate_price(total_tokens_usage: Dict[str, int], model: str) -> float:
    """Calculates the total price based on the number of promt & completion tokens and the models price for input and output tokens (per 1k tokens).

    Args:
        total_tokens_usage: OpenAI "usage" dictionaries which defines prompt_tokens, completion_tokens and total_tokens


    Returns:
        float: The price for used tokens
    """
    model_price = MODEL_PRICING[model]
    price = (total_tokens_usage["prompt_tokens"] * model_price["input"] + total_tokens_usage["completion_tokens"] * model_price["output"]) / 1000
    return price

# %% ../nbs/CLI.ipynb 9
def _get_description(input_path: str) -> str:
    """Reads description from te file and returns it as a string

    Args:
        input_path: Path to the file with the desription
        
    Raises:
        ValueError: If the file does not exist.

    Returns:
        The description string which was read from the file.
    """
    try:
        with open(input_path) as file:
            # Read all lines to list
            lines = file.readlines()
            # Join the lines 
            description = '\r'.join(lines)
            logger.info(f"Reading application description from '{str(Path(input_path).absolute())}'.")
    except Exception as e:
        raise ValueError(f"Error while reading from the file: '{str(Path(input_path).absolute())}'\n{str(e)}")
    return description

# %% ../nbs/CLI.ipynb 12
def _validate_app_description(
    description: Optional[str],
    input_path: str,
    model: OpenAIModel,
    tokens_list: List[Dict[str, int]],
) -> Tuple[str, List[Dict[str, int]]]:
    if not description:
        if not input_path:
            raise ValueError(EMPTY_DESCRIPTION_ERROR)
        description = _get_description(input_path)

    cleaned_description = strip_white_spaces(description)
    return validate_app_description(cleaned_description, model.value, tokens_list)

# %% ../nbs/CLI.ipynb 16
@app.command(
    "generate",
    help="Effortlessly create a new FastStream project based on the app description.",
)
@set_logger_level
def generate_fastkafka_app(
    description: Optional[str] = typer.Argument(
        None,
        help="""Summarize your FastStream application in a few sentences!


\nInclude details about messages, topics, servers, and a brief overview of the intended business logic.


\nThe simpler and more specific the app description is, the better the generated app will be. Please refer to the below example for inspiration:


\nCreate a FastStream application using localhost broker for testing and use the default port number. 
It should consume messages from the "input_data" topic, where each message is a JSON encoded object containing a single attribute: 'data'. 
For each consumed message, create a new message object and increment the value of the data attribute by 1. Finally, send the modified message to the 'output_data' topic.
\n""",
    ),
    input_path: str = typer.Option(
        None,
        "--input_file",
        "-i",
        help="""
        The path to the file with the app desription. This path should be relative to the current working directory.
        
        \n\nIf the app description is passed via both a --input_file and a command line argument, the description from the command line will be used to create the application.
        """,
    ),
    output_path: str = typer.Option(
        ".",
        "--output_path",
        "-o",
        help="The path to the output directory where the generated project files will be saved. This path should be relative to the current working directory.",
    ),
    model: OpenAIModel = typer.Option(
        OpenAIModel.gpt3.value,
        "--model",
        "-m",
        help=f"The OpenAI model that will be used to create the FastStream project. For better results, we recommend using '{OpenAIModel.gpt4.value}'.",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Enable verbose logging by setting the logger level to INFO.",
    ),
    save_log_files: bool = typer.Option(
        False,
        "--dev",
        "-d",
        help="Save the complete logs generated by faststream-gen inside the output_path directory.",
    ),
) -> None:
    """Effortlessly create a new FastStream project based on the app description."""
    logger.info("Project generation started.")
    try:
        tokens_list: List[Dict[str, int]] = []
        ensure_openai_api_key_set()

        # Step 1: Validate description
        validated_description, tokens_list = _validate_app_description(
            description, input_path, model, tokens_list
        )

        # Step 2: Project creation
        create_project(output_path)

        # Step 3: Get relevant application examples
        prompt_examples = get_relevant_prompt_examples(validated_description)

        # Step 4: Generate application skeleton
        tokens_list, is_valid_skeleton_code = generate_app_skeleton(
            validated_description,
            output_path,
            model.value,
            tokens_list,
            prompt_examples["description_to_skeleton"],
        )
        if is_valid_skeleton_code:
            # Step 5: Generate application and test code only if previous step is successful
            tokens_list, is_valid_app_code = generate_app_and_test(
                validated_description,
                model.value,
                output_path,
                tokens_list,
                prompt_examples["skeleton_to_app_and_test"],
            )
            if is_valid_app_code:
                # Step 5: Generate application and test code only if previous step is successful
                (
                    tokens_list,
                    is_requirements_file_valid,
                ) = fix_requirements_and_run_tests(
                    output_path, model.value, tokens_list
                )

        if not is_valid_skeleton_code:
            is_valid_app_code = False
            typer.secho(" ✘ Error: Failed to generate a valid application and test code.", fg=typer.colors.RED)
            typer.secho(" ✘ Error: Integration tests failed.", fg=typer.colors.RED)

        elif not is_valid_app_code:
            typer.secho(" ✘ Error: Integration tests failed.", fg=typer.colors.RED)

    except (ValueError, KeyError) as e:
        fg = typer.colors.RED
        typer.secho(e, err=True, fg=fg)
        raise typer.Exit(code=1)
    except Exception as e:
        fg = typer.colors.RED
        typer.secho(f"Unexpected internal error: {e}", err=True, fg=fg)
        raise typer.Exit(code=1)
    finally:
        total_tokens_usage = add_tokens_usage(tokens_list)
        price = _calculate_price(total_tokens_usage, model.value)

        fg = typer.colors.CYAN
        typer.secho(f" Tokens used: {total_tokens_usage['total_tokens']}", fg=fg)
        logger.info(f"Prompt Tokens: {total_tokens_usage['prompt_tokens']}")
        logger.info(f"Completion Tokens: {total_tokens_usage['completion_tokens']}")
        typer.secho(f" Total Cost (USD): ${round(price, 5)}", fg=fg)

    if (
        (not is_valid_skeleton_code)
        or (not is_valid_app_code)
        or (not is_requirements_file_valid)
    ):
        if output_path == ".":
            test_cmd = "pytest"
            logs_dir = LOGS_DIR_NAME
        else:
            test_cmd = f"cd {output_path} && pytest"
            logs_dir = f"{output_path}/{LOGS_DIR_NAME}"
        typer.secho(
            f"""\n\n{INCOMPLETE_APP_ERROR_MSG}

{test_cmd}

For in-depth debugging, check the {logs_dir} directory for complete logs, including individual step information.
""",
            fg=typer.colors.RED,
        )
    else:
        if not save_log_files:
            shutil.rmtree(f"{output_path}/{LOGS_DIR_NAME}")
        typer.secho("✨  All files were successfully generated!", fg=fg)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Benchmark_CLI.ipynb.

# %% auto 0
__all__ = ['BENCHMARK_RESULTS_DIR', 'app', 'benchmark']

# %% ../../nbs/Benchmark_CLI.ipynb 1
from typing import *
import shutil
from tempfile import TemporaryDirectory
from pathlib import Path
from contextlib import contextmanager
import os
import subprocess # nosec: B404: Consider possible security implications associated with the subprocess module.

import typer
from yaspin import yaspin

import numpy as np

from .._code_generator.constants import INCOMPLETE_APP_ERROR_MSG


# %% ../../nbs/Benchmark_CLI.ipynb 3
BENCHMARK_RESULTS_DIR = "benchmark-results"

# %% ../../nbs/Benchmark_CLI.ipynb 4
@contextmanager
def _set_cwd(cwd_path: Union[Path, str]) -> Generator:
    """Set the current working directory for the duration of the context manager.

    Args:
        cwd_path: The path to the new working directory.

    !!! note

        The above docstring is autogenerated by docstring-gen library (https://github.com/airtai/docstring-gen)
    """
    cwd_path = Path(cwd_path)
    original_cwd = os.getcwd()
    os.chdir(cwd_path)

    try:
        yield
    finally:
        os.chdir(original_cwd)

# %% ../../nbs/Benchmark_CLI.ipynb 5
app = typer.Typer(
    short_help="Run benchmark against pre-defined example app descriptions",
)

# %% ../../nbs/Benchmark_CLI.ipynb 6
@app.command(
    "run_benchmark",
    help="Run benchmark against pre-defined example app descriptions",
)
def benchmark(
    fixtures_path: str = typer.Argument(
        ...,
        help="The path to the pre-defined example app descriptions",
    ),
    no_repeat: int = typer.Option(
        1,
        "--repeat",
        "-r",
        help="Number of generation repetitions per app description",
    ),
) -> None:
    fixtures_path_obj = Path(fixtures_path).resolve()

    app_descriptions = [
        filename
        for filename in fixtures_path_obj.glob("*.txt")
        if "-log" not in filename.stem
    ]
    
    if no_repeat > 1:
        app_descriptions = np.repeat(np.array(app_descriptions), no_repeat).tolist()
    
    no_of_description_files = len(app_descriptions)
    typer.secho(
        f"Total app description files: {no_of_description_files}", fg=typer.colors.CYAN
    )
    success_cnt = 0
    for i, app_description in enumerate(app_descriptions):
        with yaspin(
            text=f"{i+1}/{no_of_description_files} Generating app for: {app_description.name}",
            color="cyan",
            spinner="clock",
        ) as sp:
            with TemporaryDirectory() as d:
                with _set_cwd(d):
                    cli_command = (
                        f"faststream_gen -i {app_description} -o {d}/new-project --dev"
                    )

                    # nosemgrep: python.lang.security.audit.subprocess-shell-true.subprocess-shell-true
                    result = subprocess.run(  # nosec: B602, B603 subprocess call - check for execution of untrusted input.
                        cli_command,
                        shell=True,
                        check=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        text=True,
                    )
                    sp.text = ""
                    if INCOMPLETE_APP_ERROR_MSG in str(result.stdout):
                        sp.text = ""
                        sp.color = "red"
                        sp.ok(
                            f" ✘ Error: App generated failed for: {app_description.name}"
                        )

                    else:
                        sp.ok(
                            f" ✔ App successfully generated for: {app_description.name}"
                        )
                        success_cnt += 1

                    benchmark_results_dir = fixtures_path_obj / BENCHMARK_RESULTS_DIR
                    benchmark_results_dir.mkdir(parents=True, exist_ok=True)
                    shutil.copytree(
                        str(Path(d) / "new-project"),
                        benchmark_results_dir / Path(app_description).stem,
                        dirs_exist_ok=True,
                    )

    success_rate = success_cnt / no_of_description_files
    typer.secho(
        f"Success rate: {success_rate}",
        fg=typer.colors.CYAN,
    )
    if int(success_rate) != 1:
        typer.secho(
            f"\nTo debug, go to {str(fixtures_path_obj / BENCHMARK_RESULTS_DIR)}, select one of the failed app description directories, and run 'pytest'.",
            fg=typer.colors.RED,
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/App_And_Test_Generator.ipynb.

# %% auto 0
__all__ = ['logger', 'generate_app_and_test']

# %% ../../nbs/App_And_Test_Generator.ipynb 1
from typing import *
import time
import importlib.util
from tempfile import TemporaryDirectory
from pathlib import Path
import platform
from collections import defaultdict
import subprocess  # nosec: B404: Consider possible security implications associated with the subprocess module.

from yaspin import yaspin

from .._components.logger import get_logger
from .chat import CustomAIChat, ValidateAndFixResponse
from faststream_gen._code_generator.helper import (
    write_file_contents,
    read_file_contents,
    validate_python_code,
    retry_on_error,
    set_cwd
)
from .prompts import APP_AND_TEST_GENERATION_PROMPT
from faststream_gen._code_generator.constants import (
    APPLICATION_FILE_PATH,
    TEST_FILE_PATH,
    STEP_LOG_DIR_NAMES,
    LOGS_DIR_NAME,
)

from .constants import OpenAIModel

# %% ../../nbs/App_And_Test_Generator.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/App_And_Test_Generator.ipynb 5
_code_fix_prompt = """
Fix the provided Python code following these rules:

- Provide only valid Python code, no explanations or code tags.
- Do not modify the line starting with 'from app.application'; it must remain unchanged. Adherence to this rule is crucial.
"""

def _fix_generated_code(s: str) -> str:
    ai = CustomAIChat(
        params={
            "temperature": 0.2,
        },
        model=OpenAIModel.gpt3.value,
        user_prompt=_code_fix_prompt,
    )
    response, usage = ai(s) # todo: add this usage to total usage
    return str(response)

# %% ../../nbs/App_And_Test_Generator.ipynb 7
def _split_app_and_test_code(response: str) -> Tuple[str, str]:
    app_code, test_code = response.split("### application.py ###")[1].split(
        "### test.py ###"
    )
    return app_code, test_code


def _validate_response(
    response: str, output_directory: str, **kwargs: Dict[str, Any]
) -> Tuple[List[str], str]:
    try:
        app_code, test_code = _split_app_and_test_code(response)
    except (IndexError, ValueError) as e:
        return (
            ["Please add ### application.py ### and ### test.py ### in your response"],
            response,
        )

    app_code = app_code.replace("### application.py ###", "").strip()
    test_code = test_code.strip()

    fixed_app_code = _fix_generated_code(app_code)
    fixed_test_code = _fix_generated_code(test_code).replace("from application import ", "from app.application import ")

    app_file_name = Path(output_directory) / APPLICATION_FILE_PATH
    test_file_name = Path(output_directory) / TEST_FILE_PATH

    write_file_contents(str(app_file_name), fixed_app_code)
    write_file_contents(str(test_file_name), fixed_test_code)

    with set_cwd(output_directory):
        cmd = ["pytest", "--tb=short"]
        # nosemgrep: python.lang.security.audit.subprocess-shell-true.subprocess-shell-true
        p = subprocess.run(  # nosec: B602, B603 subprocess call - check for execution of untrusted input.
            cmd,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            shell=True if platform.system() == "Windows" else False,
        )
    if p.returncode != 0:
        response = f"### application.py ###\n{fixed_app_code}\n\n### test.py ###\n{fixed_test_code}\n"
        return ([str(p.stdout.decode("utf-8"))], response)

    return ([], "")

# %% ../../nbs/App_And_Test_Generator.ipynb 11
@retry_on_error()  # type: ignore
def _generate(
    model: str,
    prompt: str,
    app_skeleton: str,
    total_usage: List[Dict[str, int]],
    output_directory: str,
    **kwargs,
) -> Tuple[str, List[Dict[str, int]], bool]:
    test_generator = CustomAIChat(
        params={
            "temperature": 0.2,
        },
        model=model,
        user_prompt=prompt,
    )
    test_validator = ValidateAndFixResponse(test_generator, _validate_response)
    validator_result = test_validator.fix(
        app_skeleton,
        total_usage,
        STEP_LOG_DIR_NAMES["app"],
        str(output_directory),
        **kwargs,
    )
    
    return (
        (validator_result, True) # type: ignore
        if isinstance(validator_result[-1], defaultdict)
        else validator_result
    )

# %% ../../nbs/App_And_Test_Generator.ipynb 14
def generate_app_and_test(
    description: str,
    model: str,
    output_directory: str,
    total_usage: List[Dict[str, int]],
    relevant_prompt_examples: str,
) -> Tuple[List[Dict[str, int]], bool]:
    """Generate integration test for the FastStream app

    Args:
        description: Validated User application description
        code_gen_directory: The directory containing the generated files.
        relevant_prompt_examples: Relevant examples to add in the prompts.

    Returns:
        The generated integration test code for the application
    """
    logger.info("==== Skeleton to App and Test Generation ====")
    with yaspin(
        text="Generating application and tests (usually takes around 30 to 90 seconds)...",
        color="cyan",
        spinner="clock",
    ) as sp:
        app_skeleton_file_name = Path(output_directory) / APPLICATION_FILE_PATH
        app_skeleton = read_file_contents(str(app_skeleton_file_name))

        prompt = (
            APP_AND_TEST_GENERATION_PROMPT.replace(
                "==== REPLACE WITH APP DESCRIPTION ====", description
            )
            .replace("==== RELEVANT EXAMPLES GOES HERE ====", relevant_prompt_examples)
            .replace("from .app import", "from app.application import")
        )

        total_usage, is_valid_app_code = _generate(
            model, prompt, app_skeleton, total_usage, output_directory
        )
        
        sp.text = ""
        if is_valid_app_code:
            message = " ✔ The application and the test files are generated."
        else:
            message = " ✘ Error: Failed to generate a valid application and test code."
            sp.color = "red"

        sp.ok(message)

        return total_usage, is_valid_app_code
